@@ -19,17 +19,14 @@
 #include "gnome-ppp.h"
 
 
-#define ACCOUNT_FILE_EXTENTION     "ppp"
-#define ACCOUNT_SHARE_PATH         "/etc/ppp-accounts"
-
-
 /* globals */
 GList *__account_list = NULL;
+GList *__account_list_watchers = NULL;
 
 
 /* prototypes */
 static GList   *duplicate_g_string_list(GList *list);
-static GList   *free_g_string_list(GList **list);
+static void     free_g_string_list(GList **list);
 static void     save_account(Account *account);
 static void     load_account();
 static gchar   *get_string_blank_default(gchar *key);
@@ -38,6 +35,8 @@ ScriptEntry  *script_entry_new(ScriptEntryType type, gchar *text);
 ScriptEntry  *script_entry_clone(ScriptEntry *script_entry);
 void          script_entry_free(ScriptEntry *script_entry);
 
+static void   broadcast_watchers();
+
 
 void
 __init_account()
@@ -59,6 +58,9 @@ account_new()
   account = g_malloc(sizeof(Account));
 
   account->speed                = 115200;
+  account->default_route        = TRUE;
+  account->lock_device          = TRUE;
+  account->locked               = FALSE;
 
   account->name                 = g_string_new("");
   account->modem_device         = g_string_new("/dev/modem");
@@ -68,25 +70,38 @@ account_new()
   account->ip                   = g_string_new("");
   account->mask                 = g_string_new("");
   account->remote               = g_string_new("");
+  account->modem_init           = g_string_new("ATZ");
 
   account->phone_list           = NULL;
   account->dns_list             = NULL;
   account->domain_list          = NULL;
   account->script_list          = NULL;
 
-  /* add new account to the account list */
-  __account_list = g_list_append(__account_list, account);
-
   return account;
 }
 
 
 void
-account_free (Account *account)
+account_add(Account *account)
 {
-  /* remove account from the account list */
-  __account_list = g_list_remove(__account_list, account);
+  if (g_list_find(__account_list, account))
+    {
+      g_error("account_add(): re-adding existing account");
+    }
 
+  __account_list = g_list_append(__account_list, account);
+  broadcast_watchers();
+}
+
+
+gint
+account_free(Account *account)
+{
+  if (g_list_find(__account_list, account))
+    {
+      g_error("account_free(): tring to free active account");
+    }
+  
   g_string_free(account->name, TRUE);
   g_string_free(account->user, TRUE);
   g_string_free(account->passwd, TRUE);
@@ -94,6 +109,7 @@ account_free (Account *account)
   g_string_free(account->ip, TRUE);
   g_string_free(account->mask, TRUE);
   g_string_free(account->remote, TRUE);
+  g_string_free(account->modem_init, TRUE);
 
   free_g_string_list(&account->phone_list);
   free_g_string_list(&account->dns_list);
@@ -102,6 +118,27 @@ account_free (Account *account)
   account_script_list_clear(account);
 
   g_free(account);
+  return 1;
+}
+
+
+gint
+account_delete(Account *account)
+{
+  if (account->locked)
+    {
+      return 0;
+    }
+
+  if (!g_list_find(__account_list, account))
+    {
+      g_error("account_delete(): deleting non-added account");
+    }
+
+  /* remove account from the account list */
+  __account_list = g_list_remove(__account_list, account);
+  account_free(account);
+  broadcast_watchers();
 }
 
 
@@ -115,16 +152,19 @@ account_clone(Account *account)
   new_account = account_new();
 
   new_account->speed               = account->speed;
+  new_account->default_route       = account->default_route;
+  new_account->lock_device         = account->lock_device;
 
-  new_account->name                = g_string_new(account->name->str);
-  new_account->modem_device        = g_string_new(account->modem_device->str);
-  new_account->user                = g_string_new(account->user->str);
-  new_account->passwd              = g_string_new(account->passwd->str);
-  new_account->remotename          = g_string_new(account->remotename->str);
-  new_account->ip                  = g_string_new(account->ip->str);
-  new_account->mask                = g_string_new(account->mask->str);
-  new_account->remote              = g_string_new(account->remote->str);
-
+  g_string_assign(new_account->name,          account->name->str);
+  g_string_assign(new_account->modem_device,  account->modem_device->str);
+  g_string_assign(new_account->user,          account->user->str);
+  g_string_assign(new_account->passwd,        account->passwd->str);
+  g_string_assign(new_account->remotename,    account->remotename->str);
+  g_string_assign(new_account->ip,            account->ip->str);
+  g_string_assign(new_account->mask,          account->mask->str);
+  g_string_assign(new_account->remote,        account->remote->str);
+  g_string_assign(new_account->modem_init,    account->modem_init->str);
+  
   new_account->phone_list          = duplicate_g_string_list(account->phone_list);
   new_account->dns_list            = duplicate_g_string_list(account->dns_list);
   new_account->domain_list         = duplicate_g_string_list(account->domain_list);
@@ -135,6 +175,7 @@ account_clone(Account *account)
     {
      script_entry = (ScriptEntry *) list->data;
      list = list->next;
+
      new_account->script_list = 
        g_list_append(new_account->script_list, script_entry_clone(script_entry));
     }
@@ -150,6 +191,36 @@ account_list()
 }
 
 
+void
+account_add_watcher(AccountWatcherCBFunc func)
+{
+  __account_list_watchers = g_list_append(__account_list_watchers, func);
+}
+
+
+void
+account_remove_watcher(AccountWatcherCBFunc func)
+{
+  __account_list_watchers = g_list_remove(__account_list_watchers, func);
+}
+
+
+static void
+broadcast_watchers()
+{
+  GList *list;
+  AccountWatcherCBFunc func;
+
+  list = __account_list_watchers;
+  while (list)
+    {
+      func = (AccountWatcherCBFunc) list->data;
+      list = list->next;
+      (*func)();
+    }
+}
+
+
 void
 account_phone_list_clear(Account *account)
 {
@@ -271,30 +342,35 @@ void
 account_save()
 {
   gint i;
+  void *iter;
+  gchar *some_key;
   GList *list;
   GString *key = g_string_new(NULL);
   GString *value = g_string_new(NULL);
   Account *account;
 
-  /* really, truely wipe out the old config file
-   * and start fresh
-   */
-  unlink(gnome_config_get_real_path("/ppp"));
-  gnome_config_clean_file("/ppp");
-  gnome_config_sync();
+  /* drop all account sections */
+  iter = gnome_config_init_iterator_sections("/gnome-ppp");
+  while (gnome_config_iterator_next(iter, &some_key, NULL))
+    {
+      int len;
+
+      len = strlen(some_key);
+      if (len >= 9 && strcmp(some_key + len - 8, ",Account") == 0)
+	{
+	  g_string_sprintf(key, "/gnome-ppp/%s", some_key);
+	  gnome_config_clean_section(key->str);
+	}
+    }
 
   list = account_list();
-  for (i = 1; list; i++)
+  for (i = 0; list; i++)
     {
       account = (Account *) list->data;
       list = list->next;
 
-      /* clean the section */
-      g_string_sprintf(key, "/ppp/Account_%d", i);
-      gnome_config_clean_section(key->str);
-
       /* form the section string */
-      g_string_sprintf(key, "/ppp/Account_%d/", i);
+      g_string_sprintf(key, "/gnome-ppp/%d,Account/", i);
       gnome_config_push_prefix(key->str);
       save_account(account);
       gnome_config_pop_prefix();
@@ -316,16 +392,16 @@ account_load()
   GString *key = g_string_new(NULL);
   GString *value = g_string_new(NULL);
 
-  for (i = 1;; i++)
+  for (i = 0;; i++)
     {
       /* form the section string */
-      g_string_sprintf(key, "/ppp/Account_%d", i);
+      g_string_sprintf(key, "/gnome-ppp/%d,Account", i);
       if (!gnome_config_has_section(key->str))
 	{
 	  break;
 	}
 
-      g_string_sprintf(key, "/ppp/Account_%d/", i);
+      g_string_sprintf(key, "/gnome-ppp/%d,Account/", i);
 
       gnome_config_push_prefix(key->str);
       load_account();
@@ -357,7 +433,7 @@ duplicate_g_string_list(GList *list)
 }
 
 
-static GList *
+static void
 free_g_string_list(GList **list)
 {
   GList *l;
@@ -376,27 +452,41 @@ free_g_string_list(GList **list)
 }
 
 
+static void
+save_string_list(gchar *base_key, GList *list)
+{
+  int i;
+  GString *key, *value;
+
+  key = g_string_new("");
+
+  for (i = 0; list; i++)
+    {
+      value = (GString *) list->data;
+      list = list->next;
+
+      g_string_sprintf(key, "%d,%s", i, base_key);
+      gnome_config_set_string(key->str, value->str);
+    }
+
+  g_string_free(key, TRUE);
+}
+
+
 static void
 save_account(Account *account)
 {
   gint i;
   GList *list;
-  GString *key = g_string_new("");
-  GString *value = g_string_new("");
+  GString *key;
+
+  key = g_string_new("");
 
   gnome_config_set_string("name", account->name->str);
   gnome_config_set_string("modem_device", account->modem_device->str);
   gnome_config_set_int("speed", account->speed);
 
-  list = account->phone_list;
-  for (i = 1; list; i++)
-    {
-      value = (GString *) list->data;
-      list = list->next;
-
-      g_string_sprintf(key, "phone_number_%d", i);
-      gnome_config_set_string(key->str, value->str);
-    }
+  save_string_list("phone_number", account->phone_list);
 
   gnome_config_set_string("user", account->user->str);
   gnome_config_set_string("passwd", account->passwd->str);
@@ -406,29 +496,12 @@ save_account(Account *account)
   gnome_config_set_string("remote", account->remote->str);
   gnome_config_set_string("mask", account->mask->str);
 
-  list = account->dns_list;
-  for (i = 1; list; i++)
-    {
-      value = (GString *) list->data;
-      list = list->next;
-
-      g_string_sprintf(key, "dns_server_%d", i);
-      gnome_config_set_string(key->str, value->str);
-    }
-
-  list = account->domain_list;
-  for (i = 1; list; i++)
-    {
-      value = (GString *) list->data;
-      list = list->next;
-
-      g_string_sprintf(key, "domain_%d", i);
-      gnome_config_set_string(key->str, value->str);
-    }
+  save_string_list("dns_server", account->dns_list);
+  save_string_list("domain", account->domain_list);
 
   /* save script! */
   list = account->script_list;
-  for (i = 1; list; i++)
+  for (i = 0; list; i++)
     {
       ScriptEntry *script_entry;
 
@@ -438,30 +511,33 @@ save_account(Account *account)
       switch (script_entry->type)
 	{
 	case SCRIPT_ENTRY_EXPECT:
-	  g_string_sprintf(key, "script_entry_%d_expect", i);
+	  g_string_sprintf(key, "%d,SE_expect", i);
 	  gnome_config_set_string(key->str, script_entry->text->str);
 	  break;
 
 	case SCRIPT_ENTRY_SEND:
-	  g_string_sprintf(key, "script_entry_%d_send", i);
+	  g_string_sprintf(key, "%d,SE_send", i);
 	  gnome_config_set_string(key->str, script_entry->text->str);
 	  break;
 
 	case SCRIPT_ENTRY_SEND_USER:
-	  g_string_sprintf(key, "script_entry_%d_send_user", i);
+	  g_string_sprintf(key, "%d,SE_send_user", i);
 	  gnome_config_set_string(key->str, "");
 	  break;
 
 	case SCRIPT_ENTRY_SEND_PASSWORD:
-	  g_string_sprintf(key, "script_entry_%d_send_password", i);
+	  g_string_sprintf(key, "%d,SE_send_password", i);
 	  gnome_config_set_string(key->str, "");
 	  break;
 	}
     }
 
+  gnome_config_set_string("modem_init", account->modem_init->str);
+  gnome_config_set_bool("default_route", account->default_route);
+  gnome_config_set_bool("lock_device", account->lock_device);
+
  cleanup:
   g_string_free(key, TRUE);
-  g_string_free(value, TRUE);
 }
 
 
@@ -471,21 +547,21 @@ load_account()
   gint i;
   gboolean junk;
   GList *list;
-  GString *key = g_string_new("");
-  GString *value = g_string_new("");
+  GString *key;
   Account *account;
 
+  key = g_string_new("");
   account = account_new();
 
   g_string_assign(account->name, get_string_blank_default("name"));
   g_string_assign(account->modem_device, get_string_blank_default("modem_device"));
   account->speed = gnome_config_get_int("speed=15200");
 
-  for (i = 1;; i++)
+  for (i = 0;; i++)
     {
       char *c;
 
-      g_string_sprintf(key, "phone_number_%d", i);
+      g_string_sprintf(key, "%d,phone_number", i);
       if ( (c = gnome_config_get_string(key->str)) == NULL)
 	{
 	  break;
@@ -501,11 +577,11 @@ load_account()
   g_string_assign(account->remote, get_string_blank_default("remote"));
   g_string_assign(account->mask, get_string_blank_default("mask"));
 
-  for (i = 1;; i++)
+  for (i = 0;; i++)
     {
       char *c;
 
-      g_string_sprintf(key, "dns_server_%d", i);
+      g_string_sprintf(key, "%d,dns_server", i);
       if ( (c = gnome_config_get_string(key->str)) == NULL)
 	{
 	  break;
@@ -513,11 +589,11 @@ load_account()
       account_dns_list_append(account, c);
     }
 
-  for (i = 1;; i++)
+  for (i = 0;; i++)
     {
       char *c;
 
-      g_string_sprintf(key, "domain_%d", i);
+      g_string_sprintf(key, "%d,domain", i);
       if ( (c = gnome_config_get_string(key->str)) == NULL)
 	{
 	  break;
@@ -526,12 +602,12 @@ load_account()
     }
 
   /* load the login script */
-  for (i = 1;; i++)
+  for (i = 0;; i++)
     {
       gchar *c;
 
       /* search for expect entries */
-      g_string_sprintf(key, "script_entry_%d_expect", i);
+      g_string_sprintf(key, "%d,SE_expect", i);
       if ( (c = gnome_config_get_string(key->str)) != NULL)
 	{
 	  account_script_list_append(account, SCRIPT_ENTRY_EXPECT, c);
@@ -539,7 +615,7 @@ load_account()
 	}
 
       /* search for expect send */
-      g_string_sprintf(key, "script_entry_%d_send", i);
+      g_string_sprintf(key, "%d,SE_send", i);
       if ( (c = gnome_config_get_string(key->str)) != NULL)
 	{
 	  account_script_list_append(account, SCRIPT_ENTRY_SEND, c);
@@ -547,7 +623,7 @@ load_account()
 	}
 
       /* search for expect send user */
-      g_string_sprintf(key, "script_entry_%d_send_user", i);
+      g_string_sprintf(key, "%d,SE_send_user", i);
       if ( (c = gnome_config_get_string(key->str)) != NULL)
 	{
 	  account_script_list_append(account, SCRIPT_ENTRY_SEND_USER, "");
@@ -555,7 +631,7 @@ load_account()
 	}
 
       /* search for expect send password */
-      g_string_sprintf(key, "script_entry_%d_send_password", i);
+      g_string_sprintf(key, "%d,SE_send_password", i);
       if ( (c = gnome_config_get_string(key->str)) != NULL)
 	{
 	  account_script_list_append(account, SCRIPT_ENTRY_SEND_PASSWORD, "");
@@ -566,9 +642,14 @@ load_account()
       break;
     }
 
+  g_string_assign(account->modem_init, get_string_blank_default("modem_init"));
+  account->default_route = gnome_config_get_bool("default_route=TRUE");
+  account->lock_device = gnome_config_get_bool("lock_device=TRUE");
+
  cleanup:
   g_string_free(key, TRUE);
-  g_string_free(value, TRUE);
+
+  account_add(account);
 }
 
 
@@ -587,7 +668,12 @@ get_string_blank_default(gchar *key)
 
 
 
+
+
+
+
 /*** NEW SAVE STUFF W/O gnome_config ***/
+/* NOT IN USE AT ALL! -JMP */
 static void
 append_string_var(GList **line_list, gchar *variable, gchar *value)
 {
