@@ -33,11 +33,13 @@
 #define PPPD_TIMEOUT       5
 #define FIRST_STATE        0
 
+
 /* a couple of macros to make the code more readable */
 #define SET_TIMEOUT(t, s) ((t) = time(NULL) + (s))
 #define TIMEOUT(t)        ((t) < time(NULL))
 #define NEXT_STATE(s)     ((s)++)
-#define CALLBACK(c, t, m) ((*(c)->func)((c)->account, (t), (m)))
+#define CALLBACK(c, t, m) ((*(c)->func)((c)->account, (t), (m), ((c)->data)))
+
 
 /* master state */
 enum
@@ -51,15 +53,42 @@ enum
 };
 
 
+/* message indexes for callback messages */
+enum
+{
+  _MSG_NO_PPPD_EXEC_PERMISSION,
+  _MSG_MODEM_LOCKED,
+  _MSG_PPPD_FORK,
+  _MSG_CLOSING,
+  _MSG_MODEM_FD_PASS_TIMEOUT,
+  _MSG_MODEM_DETECT,
+  _MSG_MODEM_DETECT_FAILED,
+  _MSG_NO_PHONE_NUMBER,
+  _MSG_DIALING,
+  _MSG_DIAL_TIMEOUT,
+  _MSG_NO_DIALTONE,
+  _MSG_BUSY,
+  _MSG_CONNECTED,
+  _MSG_RUNNING_SCRIPT,
+  _MSG_SCRIPT_COMPLETE,
+  _MSG_PPP_CONNECTED,
+  _MSG_PPP_FAIL,
+  _MSG_LAST_MESSAGE
+};
+
+
 typedef struct _Connect
 {
   Account *account;
+
   ConnectCBFunc func;
+  gpointer data;
 
   Modem *modem;
   GString *session_buffer;
   GString *modem_buffer;
   time_t timeout;
+  time_t connect_time;
 
   /* state keepers */
   gint state;
@@ -85,10 +114,12 @@ malloc_connect()
   Connect *connect = g_malloc(sizeof(Connect));
   connect->account = NULL;
   connect->func = NULL;
+  connect->data = NULL;
   connect->modem = NULL;
   connect->session_buffer = g_string_new("");
   connect->modem_buffer = g_string_new("");
   connect->timeout = 0;
+  connect->connect_time = 0;
 
   connect->state = FIRST_STATE;
   connect->get_state = FIRST_STATE;
@@ -116,8 +147,9 @@ free_connect(Connect *connect)
 
 
 /* list of all active connections  */
-static GHashTable *__connect_hash = NULL;
-static gchar *modem_responses[] =
+static gchar           *__messages[_MSG_LAST_MESSAGE];
+static GHashTable      *__connect_hash = NULL;
+static gchar           *__modem_responses[] =
 {
   "NO DIALTONE\r",
   "NO CARRIER\r",
@@ -126,6 +158,7 @@ static gchar *modem_responses[] =
   NULL
 };
 
+
 /* state machine for the entire connection sequence */
 static gint state_engine_iteration(Connect *connect);
 static void state_get_modem(Connect *connect);
@@ -135,9 +168,11 @@ static void state_run_script(Connect *connect);
 static void state_exec_pppd(Connect *connect);
 static void state_connected(Connect *connect);
 
+
 /* callbacks */
 static void pppd_die_cb(gpointer data);
 
+
 /* misc */
 static void destroy_connect(Connect *connect);
 static gboolean read_charactor(Connect *connect);
@@ -155,12 +190,61 @@ __init_connect()
     {
       done_init = TRUE;
       __connect_hash = g_hash_table_new(g_direct_hash, NULL);
+
+      /* initalize messages */
+      __messages[_MSG_NO_PPPD_EXEC_PERMISSION] =
+	_("You don't have permission to execute pppd.");
+
+      __messages[_MSG_MODEM_LOCKED] =
+	_("The modem is in use by another program.");
+
+      __messages[_MSG_PPPD_FORK] =
+	_("Running pppd.");
+
+      __messages[_MSG_CLOSING] =
+	_("Closing connection.");
+
+      __messages[_MSG_MODEM_FD_PASS_TIMEOUT] =
+	_("The pppd program failed to run gnome-ppp-chat.");
+
+      __messages[_MSG_MODEM_DETECT] =
+	_("Initalizing modem.");
+
+      __messages[_MSG_MODEM_DETECT_FAILED] =
+	_("Your modem is not responding to initalization.");
+
+      __messages[_MSG_NO_PHONE_NUMBER] =
+	_("No phone numbers specified for this account.");
+
+      __messages[_MSG_DIALING] =
+	_("Dialing.");
+
+      __messages[_MSG_NO_DIALTONE] =
+	_("No dialtone.");
+
+      __messages[_MSG_BUSY] =
+	_("Number Busy.");
+
+      __messages[_MSG_CONNECTED] =
+	_("Connected.");
+
+      __messages[_MSG_RUNNING_SCRIPT] =
+	_("Running Script.");
+
+      __messages[_MSG_SCRIPT_COMPLETE] =
+	_("Script Complete.");
+
+      __messages[_MSG_PPP_CONNECTED] =
+	_("PPP connection established.");
+
+      __messages[_MSG_PPP_FAIL] =
+	_("Failed to establish a PPP connection.");
     }
 }
 
 
 gint
-connect_start(Account *account, ConnectCBFunc func)
+connect_start(Account *account, ConnectCBFunc func, gpointer data)
 {
   Connect *connect;
 
@@ -176,29 +260,39 @@ connect_start(Account *account, ConnectCBFunc func)
   connect = malloc_connect();
   connect->account = account;
   connect->func = func;
+  connect->data = data;
 
   /* add to the hash */
   g_hash_table_insert(__connect_hash, account, connect);
 
-  /* XXX: we want to check for an existing device lock here! */
-
   /* for pppd -- this forks pppd with the gnome-ppp-chat */
-  switch (pppd_exec(connect->account, pppd_die_cb, connect))
+  if (pppd_exec(connect->account, pppd_die_cb, connect))
     {
-    case 0:
-      CALLBACK(connect, PPP_DISCONNECTED, "Failed to fork pppd");
-      goto error;
-      break;
+      CALLBACK(connect, PPP_IN_PROGRESS, __messages[_MSG_PPPD_FORK]);
+    }
+  else
+    {
+      switch (pppd_errno)
+	{
+	case PPPD_NO_EXEC_PERMISSION:
+	  CALLBACK(connect, PPP_ERROR, __messages[_MSG_NO_PPPD_EXEC_PERMISSION]);
+	  goto error;
 
-    default:
-      CALLBACK(connect, PPP_IN_PROGRESS, "Forked pppd");
-      break;
+	case PPPD_MODEM_LOCKED:
+	  CALLBACK(connect, PPP_ERROR, __messages[_MSG_MODEM_LOCKED]);
+	  goto error;
+
+	default:
+	  g_error("connect_start(): unknown pppd error");
+	  break;
+	}
     }
 
   return 1;
 
  error:
   /* remove connection from hash */
+  CALLBACK(connect, PPP_DISCONNECTED, NULL);
   g_hash_table_remove(__connect_hash, connect->account);
   free_connect(connect);
   return 0;
@@ -235,11 +329,17 @@ connect_stop(Account *account)
       return;
     }
 
-  /* XXX: always setting the PPP_CONNECTED state here is a bit
-   * of a cheat to get the status message to go away in
-   * the GUI -- there needs to be a PPP_ERROR message
+  /* inform the user the connection is being closed, make sure
+   * to use the proper connect state in the callback
    */
-  CALLBACK(connect, PPP_CONNECTED, "Closing Connection");
+  if (connect->state == _CONNECTED)
+    {
+      CALLBACK(connect, PPP_CONNECTED, __messages[_MSG_CLOSING]);
+    }
+  else
+    {
+      CALLBACK(connect, PPP_IN_PROGRESS, __messages[_MSG_CLOSING]);
+    }
   connect->close = TRUE;
 }
 
@@ -326,45 +426,42 @@ state_get_modem(Connect *connect)
       NEXT_STATE(connect->get_state);
       break;
 
+
     case __RECIEVE_FD:
-      switch (pppd_modem_fd(connect->account, &modem_fd))
+      if (pppd_modem_fd(connect->account, &modem_fd))
 	{
-	  /* success, recieved modem fd */
-	case 1:
 	  connect->modem = modem_new(modem_fd, connect->account->speed);
 	  if (!connect->modem)
 	    {
-	      CALLBACK(connect, PPP_IN_PROGRESS, "Modem is not a TTY");
-	      goto error;
+	      g_error("state_get_modem(): ioctl failure on modem");
 	    }
 	  NEXT_STATE(connect->state);
-
-	  /* have not recieved the modem fd yet */
-	case 0:
-	  CALLBACK(connect, PPP_IN_PROGRESS, "Waiting...");
-
-	  if (TIMEOUT(connect->timeout))
-	    {
-	      CALLBACK(connect, PPP_IN_PROGRESS, "Random pppd problem");
-	      goto error;
-	    }
-	  break;
-
-	  /* there was an error with the domain socket used for
-	   * passing the modem's file descriptor
-	   */
-	case -1:
-	  CALLBACK(connect, PPP_IN_PROGRESS, "Domain socket error");
-	  goto error;
-	  break;
 	}
-      break;
+      else
+	{
+	  switch (pppd_errno)
+	    {
+	    case PPPD_PENDING:
+	      if (TIMEOUT(connect->timeout))
+		{
+		  CALLBACK(connect, PPP_ERROR, __messages[_MSG_MODEM_FD_PASS_TIMEOUT]);
+		  connect->close = TRUE;
+		  goto error;
+		}
+	      break;
+
+	    case PPPD_LOOKUP_FAIL:
+	      goto error;
+
+	    default:
+	      g_error("state_get_modem(): unknown pppd errno");
+	      break;
+	    }
+	}
     }
 
-  return;
-
  error:
-  connect->close = TRUE;
+  return;
 }
 
 
@@ -372,24 +469,30 @@ static void
 state_detect_modem(Connect *connect)
 {
   gint index;
+  GString *gstr;
 
   enum
   {
-    __PROBE,
-    __LISTEN
+    __AT_INIT,
+    __OK
   };
 
+  gstr = g_string_new("");
+
   switch (connect->detect_state)
     {
-    case __PROBE:
-      CALLBACK(connect, PPP_IN_PROGRESS, "Detecting Modem");
-      modem_write(connect->modem, "AT\r", 3);
+    case __AT_INIT:
+      CALLBACK(connect, PPP_IN_PROGRESS, __messages[_MSG_MODEM_DETECT]);
+
+      g_string_sprintf(gstr, "%s\r", connect->account->modem_init->str);
+      modem_write(connect->modem, gstr->str, gstr->len);
+
       SET_TIMEOUT(connect->timeout, DETECT_TIMEOUT);
       NEXT_STATE(connect->detect_state);
       break;
       
 
-    case __LISTEN:
+    case __OK:
       index = g_string_find(connect->modem_buffer, "OK\r", TRUE);
 
       /* modem deteted -- flush the modem buffers and go to the
@@ -397,25 +500,25 @@ state_detect_modem(Connect *connect)
        */
       if (search_modem_buffer(connect, "OK\r"))
 	{
-	  CALLBACK(connect, PPP_IN_PROGRESS, "Modem Found");
 	  modem_drain(connect->modem);
 	  NEXT_STATE(connect->state);
-	  goto cleanup;
 	}
-
-      /* we've timed out while trying to detect the modem, inform
-       * the GUI and set the close flag for this connection
-       */
-      if (TIMEOUT(connect->timeout))
+      else
 	{
-	  CALLBACK(connect, PPP_IN_PROGRESS, "Modem Not Found!");
-	  connect->close = TRUE;
+	  /* we've timed out while trying to detect the modem, inform
+	   * the GUI and set the close flag for this connection
+	   */
+	  if (TIMEOUT(connect->timeout))
+	    {
+	      CALLBACK(connect, PPP_ERROR, __messages[_MSG_MODEM_DETECT_FAILED]);
+	      connect->close = TRUE;
+	    }
 	}
       break;
     }
 
  cleanup:
-  return;
+  g_string_free(gstr, TRUE);
 }
 
 
@@ -441,7 +544,7 @@ state_dial(Connect *connect)
       /* the account HAS to have at least one phone number to dial */
       if (g_list_length(connect->account->phone_list) == 0)
 	{
-	  CALLBACK(connect, PPP_IN_PROGRESS, "No Phone Numbers Specified");
+	  CALLBACK(connect, PPP_ERROR, __messages[_MSG_NO_PHONE_NUMBER]);
 	  connect->close = TRUE;
 	  goto cleanup;
 	}
@@ -460,8 +563,7 @@ state_dial(Connect *connect)
       phone_number = (GString *) connect->phone_number_node->data;
       if (phone_number)
 	{
-	  g_string_sprintf(gstr, "Dialing %s", phone_number->str);
-	  CALLBACK(connect, PPP_IN_PROGRESS, gstr->str);
+	  CALLBACK(connect, PPP_IN_PROGRESS, __messages[_MSG_DIALING]);
 
 	  modem_drain(connect->modem);
 
@@ -474,12 +576,11 @@ state_dial(Connect *connect)
 
 
     case __ANSWER:
-      result = search_modem_buffer_list(connect, modem_responses);
+      result = search_modem_buffer_list(connect, __modem_responses);
 
       /* TIMEOUT */
       if (!result && TIMEOUT(connect->timeout))
 	{
-	  CALLBACK(connect, PPP_IN_PROGRESS, "Timeout");
 	  g_string_assign(connect->modem_buffer, "");
 	  connect->dial_state = __CALL;
 	  goto cleanup;
@@ -494,7 +595,7 @@ state_dial(Connect *connect)
       /* no dialtone */
       if (strcmp(result, "NO DIALTONE\r") == 0)
 	{
-	  CALLBACK(connect, PPP_IN_PROGRESS, "Phone Line Not Plugged Into Modem");
+	  CALLBACK(connect, PPP_ERROR, __messages[_MSG_NO_DIALTONE]);
 	  connect->close = TRUE;
 	  goto cleanup;
 	}
@@ -502,7 +603,7 @@ state_dial(Connect *connect)
       /* no carrior */
       if (strcmp(result, "NO CARRIER\r") == 0)
 	{
-	  CALLBACK(connect, PPP_IN_PROGRESS, "Number Busy");
+	  CALLBACK(connect, PPP_IN_PROGRESS, __messages[_MSG_BUSY]);
 	  connect->dial_state = __CALL;
 	  goto cleanup;
 	}
@@ -510,7 +611,7 @@ state_dial(Connect *connect)
       /* busy */
       if (strcmp(result, "BUSY\r") == 0)
 	{
-	  CALLBACK(connect, PPP_IN_PROGRESS, "Number Busy");
+	  CALLBACK(connect, PPP_IN_PROGRESS, __messages[_MSG_BUSY]);
 	  connect->dial_state = __CALL;
 	  goto cleanup;
 	}
@@ -518,7 +619,7 @@ state_dial(Connect *connect)
       /* we've connected */
       if (strcmp(result, "CONNECT") == 0)
 	{
-	  CALLBACK(connect, PPP_IN_PROGRESS, "Connected");
+	  CALLBACK(connect, PPP_IN_PROGRESS, __messages[_MSG_CONNECTED]);
 	  NEXT_STATE(connect->state);
 	  goto cleanup;
 	}
@@ -544,6 +645,8 @@ state_run_script(Connect *connect)
    */
   if (!connect->script_entry_node)
     {
+      CALLBACK(connect, PPP_IN_PROGRESS, __messages[_MSG_RUNNING_SCRIPT]);
+
       if (connect->account->script_list)
 	{
 	  connect->script_entry_node = connect->account->script_list;
@@ -564,36 +667,28 @@ state_run_script(Connect *connect)
   switch (script_entry->type)
     {
     case SCRIPT_ENTRY_EXPECT:
-      g_string_sprintf(gstr, "Expecting: %s", script_entry->text->str);
-      CALLBACK(connect, PPP_IN_PROGRESS, gstr->str);
-
       if (search_modem_buffer(connect, script_entry->text->str))
 	{
 	  next = TRUE;
 	}
       break;
 
-    case SCRIPT_ENTRY_SEND:
-      g_string_sprintf(gstr, "Sending: %s", script_entry->text->str);
-      CALLBACK(connect, PPP_IN_PROGRESS, gstr->str);
 
+    case SCRIPT_ENTRY_SEND:
       g_string_sprintf(gstr, "%s\r\n", script_entry->text->str);
       modem_write(connect->modem, gstr->str, gstr->len);
       next = TRUE;
       break;
 
-    case SCRIPT_ENTRY_SEND_USER:
-      g_string_sprintf(gstr, "Sending User Name: %s", connect->account->user->str);
-      CALLBACK(connect, PPP_IN_PROGRESS, gstr->str);
 
+    case SCRIPT_ENTRY_SEND_USER:
       g_string_sprintf(gstr, "%s\r\n", connect->account->user->str);
       modem_write(connect->modem, gstr->str, gstr->len);
       next = TRUE;
       break;
+
       
     case SCRIPT_ENTRY_SEND_PASSWORD:
-      CALLBACK(connect, PPP_IN_PROGRESS, "Sending Password");
-
       g_string_sprintf(gstr, "%s\r\n", connect->account->passwd->str);
       modem_write(connect->modem, gstr->str, gstr->len);
       next = TRUE;
@@ -610,7 +705,7 @@ state_run_script(Connect *connect)
  
       if (connect->script_entry_node == NULL)
 	{
-	  CALLBACK(connect, PPP_IN_PROGRESS, "Script Complete");
+	  CALLBACK(connect, PPP_IN_PROGRESS, __messages[_MSG_SCRIPT_COMPLETE]);
 	  NEXT_STATE(connect->state);
 	}
     }
@@ -624,21 +719,17 @@ static void
 state_exec_pppd(Connect *connect)
 {
   gint device_number;
-  GString *gstr;
 
   enum
   {
     __START_PPP,
-    __DETECT_SUCESSFUL_PPP_CONNECTION
+    __DETECT_PPP_CONNECTION
   };
   
-  gstr = g_string_new("");
 
   switch (connect->pppd_state)
     {
     case __START_PPP:
-      CALLBACK(connect, PPP_IN_PROGRESS, "Starting PPP");
-      
       /* close the connection to the modem */
       modem_destroy(connect->modem);
       connect->modem = NULL;
@@ -646,37 +737,86 @@ state_exec_pppd(Connect *connect)
       /* ends gnome-ppp-chat allowing pppd to attempt to 
        * establish a PPP connection
        */
-      pppd_end_chat(connect->account);
-      
-      SET_TIMEOUT(connect->timeout, PPPD_TIMEOUT);
-      NEXT_STATE(connect->pppd_state);
-      break;
-      
-    case __DETECT_SUCESSFUL_PPP_CONNECTION:
-      device_number = pppd_device_number(connect->account);
-      if (device_number >= 0)
+      if (pppd_end_chat(connect->account))
 	{
-	  g_string_sprintf(gstr, "Connected on device ppp%d", device_number);
-	  CALLBACK(connect, PPP_CONNECTED, gstr->str);
+	  SET_TIMEOUT(connect->timeout, PPPD_TIMEOUT);
+	  NEXT_STATE(connect->pppd_state);
+	}
+      else
+	{
+	  switch (pppd_errno)
+	    {
+	    case PPPD_LOOKUP_FAIL:
+	      goto error;
+ 
+	    default:
+	      g_error("state_exec_pppd(): unknown error from pppd_end_chat");
+	      break;
+	    }
+	}
+      break;
+
+
+    case __DETECT_PPP_CONNECTION:
+      if (pppd_device_number(connect->account, &device_number))
+	{
+	  CALLBACK(connect, PPP_CONNECTED, __messages[_MSG_PPP_CONNECTED]);
 	  NEXT_STATE(connect->state);
 	}
-
-      if (TIMEOUT(connect->timeout))
+      else
 	{
-	  CALLBACK(connect, PPP_IN_PROGRESS, "Failure to establish PPP link");
-	  connect->close = TRUE;
+	  switch (pppd_errno)
+	    {
+	    case PPPD_PENDING:
+	      if (TIMEOUT(connect->timeout))
+		{
+		  CALLBACK(connect, PPP_ERROR, __messages[_MSG_PPP_FAIL]);
+		  connect->close = TRUE;
+		  goto error;
+		}
+	      break;
+
+	    case PPPD_LOOKUP_FAIL:
+	      goto error;
+	      
+	    default:
+	      g_error("state_exec_pppd(): unknown error from pppd_device_number");
+	      break;
+	    }
 	}
       break;
     }
 
- cleanup:
-  g_string_free(gstr, TRUE);
+ error:
 }
 
 
 static void
 state_connected(Connect *connect)
 {
+#if 0
+  GString *gstr;
+  gint total_seconds, hours, minutes;
+
+  gstr = g_string_new("");
+
+  /* set the inital time of connect */
+  if (connect->connect_time == 0)
+    {
+      connect->connect_time = time(NULL);
+    }
+
+  total_seconds = time(NULL) - connect->connect_time;
+
+  hours = total_seconds / 3600;
+  minutes = (total_seconds % 3600) / 60;
+
+  g_string_sprintf(gstr, "Connected %d:%d", hours, minutes);
+  CALLBACK(connect, PPP_CONNECTED, gstr->str);
+
+ cleanup:
+  g_string_free(gstr, TRUE);
+#endif
 }
 
 
