@@ -35,6 +35,7 @@
 #include "gnome-ppp.h"
 #include "misc.h"
 #include "pppd.h"
+#include "dns.h"
 
 
 #if HAVE_SOCKETBITS_H
@@ -65,7 +66,6 @@
 
 /* constants */
 #define PPPD_PATH             "/usr/sbin/pppd"
-#define RESOLV_CONF_PATH      "/etc/resolv.conf"
 #define PPPD_PID_PATH         "/var/run"
 #define MAX_ARGS              25
 
@@ -117,6 +117,7 @@ free_ppp(PPP *ppp)
 
 
 /* static globals */
+gint                            pppd_errno;
 static GHashTable              *__ppp_hash = NULL;
 static GHashTable              *__ppp_pid_hash = NULL;
 static struct sigaction         sact;
@@ -128,14 +129,10 @@ static pid_t fork_pppd(PPP *ppp);
 static void handle_sigchld(int signum);
 static void block_sigchld();
 static void unblock_sigchld();
-
-static gint  check_pid_file(PPP *ppp);
+static gint check_device_lock(gchar *device_name);
+static gint check_pid_file(PPP *ppp);
 static pid_t get_pid(gint device_number);
-static gint  is_pid_filename(gchar *filename);
-
-static void add_dns_entries(PPP *ppp);
-static void remove_dns_entries(PPP *ppp);
-
+static gint is_pid_filename(gchar *filename);
 static int recieve_fd(int fd, int *recvfd);
 
 
@@ -171,21 +168,28 @@ pppd_exec(Account *account, PPPCBFunc func, gpointer data)
   struct sockaddr_un name;
   PPP *ppp;
 
-  g_assert(account != NULL);
-
   ppp = NULL;
 
+  g_assert(account != NULL);
+  block_sigchld();
+  if (g_hash_table_lookup(__ppp_hash, account))
+    {
+      g_error("pppd_exec(): account connection in progress");
+    }
+
   /* check to see if the process has permission to
    * execute the pppd daemon
    */
   if (access(PPPD_PATH, X_OK) < 0)
     {
+      pppd_errno = PPPD_NO_EXEC_PERMISSION; 
       goto error;
     }
 
-  block_sigchld();
-  if (g_hash_table_lookup(__ppp_hash, account))
+  /* check to see if the dialout device is locked */
+  if (check_device_lock(account->modem_device->str))
     {
+      pppd_errno = PPPD_MODEM_LOCKED;
       goto error;
     }
 
@@ -201,7 +205,6 @@ pppd_exec(Account *account, PPPCBFunc func, gpointer data)
   if (mktemp(ppp->socket_path->str) == NULL)
     {
       g_error("pppd_exec(): mktemp");
-      goto error;
     }
 
   unlink(ppp->socket_path->str);
@@ -214,13 +217,11 @@ pppd_exec(Account *account, PPPCBFunc func, gpointer data)
   if ((ppp->socket_fd = socket(AF_UNIX, SOCK_STREAM, 0)) < 0)
     {
       g_error("pppd_exec(): socket");
-      goto error;
     }
 
   if (bind(ppp->socket_fd, (struct sockaddr *) &name, len) < 0)
     {
       g_error("pppd_exec(): bind");
-      goto error;
     }
 
   /* fork/exec pppd
@@ -232,7 +233,6 @@ pppd_exec(Account *account, PPPCBFunc func, gpointer data)
   if (ppp->pid <= 0)
     {
       g_error("pppd_exec(): fork_pppd <= 0");
-      goto error;
     }
   g_hash_table_insert(__ppp_pid_hash, &ppp->pid, ppp);
   unblock_sigchld();
@@ -269,17 +269,18 @@ pppd_modem_fd(Account *account, int *modem_fd)
   PPP *ppp;
 
   g_assert(account != NULL);
-
-  /* set to invalid file descriptor */
-  *modem_fd = -1;
-
   block_sigchld();
   ppp = g_hash_table_lookup(__ppp_hash, account);
   if (!ppp)
     {
-      goto error;
+      pppd_errno = PPPD_LOOKUP_FAIL;
+      unblock_sigchld();
+      return 0;
     }
 
+  /* set to invalid file descriptor */
+  *modem_fd = -1;
+
   FD_ZERO(&rfds);
   FD_SET(ppp->socket_fd, &rfds);
   tv.tv_sec = 0;
@@ -297,50 +298,48 @@ pppd_modem_fd(Account *account, int *modem_fd)
       
       if (listen(ppp->socket_fd, 5) < 0)
 	{
-	  goto error;
+	  g_error("pppd_modem_fd(): listen");
 	}
       
       ppp->client_fd = accept(ppp->socket_fd, (struct sockaddr *) &name, &len);
       if (ppp->client_fd < 0)
 	{
-	  goto error;
+	  g_error("pppd_modem_fd(): accept");
 	}
       
       if (!recieve_fd(ppp->client_fd, modem_fd))
 	{
-	  goto error;
+	  g_error("pppd_modem_fd(): recieve_fd");
 	}
     }
   else
     {
-      goto pending;
+      pppd_errno = PPPD_PENDING;
+      goto error;
     }
       
   unblock_sigchld();
   return 1;
 
- pending:
-  unblock_sigchld();
-  return 0;
-
  error:
   unblock_sigchld();
-  return -1;
+  return 0;
 }
 
 
-void
+gint
 pppd_end_chat(Account *account) 
 {
   PPP *ppp;
 
   g_assert(account != NULL);
-
   block_sigchld();
   ppp = g_hash_table_lookup(__ppp_hash, account);
   if (!ppp) 
     {
-      goto cleanup;
+      pppd_errno = PPPD_LOOKUP_FAIL;
+      unblock_sigchld();
+      return 0;
     }
 
   /* signal gnome-ppp-chat to shut down */
@@ -351,24 +350,24 @@ pppd_end_chat(Account *account)
       ppp->client_fd = -1;
     }
 
- cleanup:
   unblock_sigchld();
+  return 1;
 }
 
 
 gint
-pppd_device_number(Account *account)
+pppd_device_number(Account *account, gint *device_number)
 {
   PPP *ppp;
-  gint device_number;
 
   g_assert(account != NULL);
-
   block_sigchld();
   ppp = g_hash_table_lookup(__ppp_hash, account);
   if (!ppp) 
     {
-      goto error;
+      pppd_errno = PPPD_LOOKUP_FAIL;
+      unblock_sigchld();
+      return 0;
     }
 
   /* if the ppp device number has not been determined, then
@@ -379,21 +378,21 @@ pppd_device_number(Account *account)
       ppp->device_number = check_pid_file(ppp);
       if (ppp->device_number >= 0)
 	{
-	  add_dns_entries(ppp);
+	  dns_add_entries(ppp->account, ppp->pid);
 	}
    } 
-  device_number = ppp->device_number;
+  *device_number = ppp->device_number;
  
   unblock_sigchld();
-  return device_number;
+  return 1;
 
  error:
   unblock_sigchld();
-  return -1;
+  return 0;
 }
 
 
-void
+gint
 pppd_kill(Account *account) 
 {
   PPP *ppp;
@@ -404,7 +403,9 @@ pppd_kill(Account *account)
   ppp = g_hash_table_lookup(__ppp_hash, account);
   if (!ppp)
     {
-      goto error;
+      pppd_errno = PPPD_LOOKUP_FAIL;
+      unblock_sigchld();
+      return 0;
     }
 
   /* sanity checks */
@@ -414,9 +415,8 @@ pppd_kill(Account *account)
     }
 
   kill(ppp->pid, SIGTERM);
-
- error:
   unblock_sigchld();
+  return 1;
 }
 
 
@@ -427,22 +427,19 @@ pppd_pid(Account *account)
   pid_t pid;
 
   g_assert(account != NULL);
-
   block_sigchld();
   ppp = g_hash_table_lookup(__ppp_hash, account);
-  if (!ppp) 
+  if (!ppp)
     {
-      goto error;
+      pppd_errno = PPPD_LOOKUP_FAIL;
+      unblock_sigchld();
+      return 0;
     }
 
   pid = ppp->pid;
 
   unblock_sigchld();
   return pid;
-
- error:
-  unblock_sigchld();
-  return 0;
 }
 
 
@@ -452,7 +449,7 @@ fork_pppd(PPP *ppp)
   pid_t pid;
   int arg_index;
   char *args[MAX_ARGS];
-  GString *device, *speed, *user, *remotename, *chat;
+  GString *device, *speed, *user, *remotename, *chat, *netmask, *ip;
 
   arg_index = 0;
 
@@ -461,6 +458,8 @@ fork_pppd(PPP *ppp)
   user = g_string_new("");
   remotename = g_string_new("");
   chat = g_string_new("");
+  netmask = g_string_new("");
+  ip = g_string_new("");
 
   /* name of the program as argv[0] */
   args[arg_index++] = "pppd";
@@ -474,17 +473,54 @@ fork_pppd(PPP *ppp)
   args[arg_index++] = speed->str;
   
   /* local:remote ip address */
+  if (ppp->account->ip->len > 0 || ppp->account->remote->len > 0)
+    {
+      if (ppp->account->ip->len > 0)
+	{
+	  g_string_assign(ip, ppp->account->ip->str);
+	}
+      else
+	{
+	  g_string_assign(ip, "0.0.0.0");
+	}
+
+      g_string_append_c(ip, ':');
+
+      if (ppp->account->remote->len > 0)
+      	{
+	  g_string_append(ip, ppp->account->remote->str);
+	}
+      else
+	{
+	  g_string_append(ip, "0.0.0.0");
+	}
+
+      args[arg_index++] = ip->str;
+    }
+
   /* netmask */
+  if (ppp->account->mask->len > 0)
+    {
+      args[arg_index++] = "netmask";
+      g_string_sprintf(netmask, "%s", ppp->account->mask->str);
+      args[arg_index++] = netmask->str;
+    }
   
   /* lock the device */
-  args[arg_index++] = "lock";
-  
+  if (ppp->account->lock_device)
+    {
+      args[arg_index++] = "lock";
+    }
+
   /* don't fork off */
   args[arg_index++] = "-detach";
   
   /* add default route */
-  args[arg_index++] = "defaultroute";
-  
+  if (ppp->account->default_route)
+    {
+      args[arg_index++] = "defaultroute";
+    }
+
   /* use CRTSCTS flow control */
   args[arg_index++] = "crtscts";
   
@@ -544,6 +580,8 @@ fork_pppd(PPP *ppp)
   g_string_free(user, TRUE);
   g_string_free(remotename, TRUE);
   g_string_free(chat, TRUE);
+  g_string_free(netmask, TRUE);
+  g_string_free(ip, TRUE);
 
   return pid;
 }
@@ -561,7 +599,7 @@ handle_sigchld(int signum)
   ppp = g_hash_table_lookup(__ppp_pid_hash, &pid);
   if (ppp)
     {
-      remove_dns_entries(ppp);
+      dns_remove_entries(ppp->account, ppp->pid);
 
       g_hash_table_remove(__ppp_hash, ppp->account);
       g_hash_table_remove(__ppp_pid_hash, &pid);
@@ -623,6 +661,48 @@ unblock_sigchld()
 }
 
 
+static gint
+check_device_lock(gchar *device_name)
+{
+  int fd;
+  pid_t pid;
+  char inbuf[20];
+  GString *filename;
+
+  /* create lock file path */
+  filename = g_string_new("");
+  g_string_sprintf(filename, "/var/lock/LCK..%s", g_basename(device_name));
+  fd = open(filename->str, O_RDONLY);
+  g_string_free(filename, TRUE);
+
+  if (fd >= 0)
+    {
+      /* prevent race condition */
+      sleep(1);
+
+      read(fd, inbuf, 19);
+      close(fd);
+
+      inbuf[19] = '\0';
+      sscanf(inbuf, "%d", &pid);
+
+      /* check the process ID stored in the lockfile to
+       * see if it is stale (written by dead process)
+       */
+      if(kill(pid, 0) == -1 && errno == ESRCH)
+	{
+	  return 0;
+	} 
+      else
+	{
+	  return 1;
+	}
+    }
+
+  return 0;
+}
+
+
 /*** PPP<X>.PID Run Files ***/
 
 /* returns the device number of the pppd device if a coorasponding
@@ -734,139 +814,6 @@ is_pid_filename(gchar *filename)
 }
 
 
-/*** DNS ***/
-static void
-add_dns_entries(PPP *ppp)
-{
-  GList *list, *line_list;
-  GString *gstr, *line, *search_line, *dns_line;
-
-  /* check for write permission to resolv.conf, return
-   * if we don't have it
-   */
-  if (access(RESOLV_CONF_PATH, W_OK) < 0)
-    {
-      return;
-    }
-
-  /* utility gstrings */
-  gstr = g_string_new("");
-
-  /* read the resolv.conf file
-   * FIXME: should flock the file before reading, and unlock
-   *        it after writing
-   */
-  line_list = j_read_lines(RESOLV_CONF_PATH);
-
-  /* DNS servers */
-  if (g_list_length(ppp->account->dns_list) > 0)
-    {
-      /* start from the end of the DNS list and work
-       * towards the beginning so the search ends up in
-       * order
-       */
-      list = g_list_last(ppp->account->dns_list);
-      while (list)
-	{
-	  line = (GString *) list->data;
-	  list = list->prev;
-
-	  dns_line = g_string_new("");
-	  g_string_sprintf(
-              dns_line,
-	      "nameserver %s # GNOME-PPP TEMP ENTRY PID: %d\n",
-	      line->str,
-	      ppp->pid);
-
-	  /* append it to the line_list to be written */
-	  line_list = g_list_prepend(line_list, dns_line);
-	}
-    }
-
-  /* DNS search domains */
-  if (g_list_length(ppp->account->domain_list) > 0)
-    {
-      search_line = g_string_new("search");
-
-      list = ppp->account->domain_list;
-      while (list)
-	{
-	  line = (GString *) list->data;
-	  list = list->next;
-
-	  g_string_append_c(search_line, ' ');
-	  g_string_append(search_line, line->str);
-	}
-
-      g_string_sprintf(gstr, " # GNOME-PPP TEMP ENTRY PID: %d\n", ppp->pid);
-      g_string_append(search_line, gstr->str);
-      
-      /* append it to the line_list to be written */
-      line_list = g_list_prepend(line_list, search_line);
-    }
-
-  /* write the new resolv.conf file and free the line list */
-  j_write_lines(RESOLV_CONF_PATH, line_list);
-  j_free_lines(line_list);
-  g_string_free(gstr, TRUE);
-}
-
-
-static void
-remove_dns_entries(PPP *ppp)
-{
-  GList *list, *old_line_list, *new_line_list;
-  GString *search, *line;
-
-  /* check for write permission to resolv.conf, return
-   * if we don't have it
-   */
-  if (access(RESOLV_CONF_PATH, W_OK) < 0)
-    {
-      return;
-    }
-
-  /* create the search string */
-  search = g_string_new(" # GNOME-PPP TEMP ENTRY PID:");
-  /* g_string_sprintf(search, " # GNOME-PPP TEMP ENTRY PID: %d\n", ppp->pid); */
-
-  /* read the resolv.conf file
-   * FIXME: should flock the file before reading, and unlock
-   *        it after writing
-   */
-  old_line_list = j_read_lines(RESOLV_CONF_PATH);
-  new_line_list = NULL;
-
-  /* iterate through all the lines in the resolv.conf file, freeing
-   * the temporary entries and adding the perminate ones to
-   * new_line_list to be written back to the resolv.conf file
-   */
-  list = old_line_list;
-  while (list)
-    {
-      line = (GString *) list->data;
-      list = list->next;
-
-      if (g_string_find(line, search->str, FALSE) < 0)
-	{
-	  new_line_list = g_list_append(new_line_list, line);
-	}
-      else
-	{
-	  g_string_free(line, TRUE);
-	}
-    }
-
-  /* free the old list */
-  g_list_free(old_line_list);
-  
-  /* write the new resolv.conf file and free the line list */
-  j_write_lines(RESOLV_CONF_PATH, new_line_list);
-  j_free_lines(new_line_list);
-  g_string_free(search, TRUE);
-}
-
-
 /* for file descriptor passing between gnome-ppp-chat and
  * libgnome-ppp
  */
