@@ -21,11 +21,24 @@
 
 typedef void (*PPPCBFunc)(gpointer data);
 
+enum
+{
+  PPPD_OK,
+  PPPD_LOOKUP_FAIL,
+  PPPD_PENDING,
+  PPPD_MODEM_LOCKED,
+  PPPD_NO_EXEC_PERMISSION,
+};
+
+
+extern gint pppd_errno;
+
+
 gint          pppd_exec(Account *account, PPPCBFunc func, gpointer data);
 gint          pppd_modem_fd(Account *account, int *modem_fd);
-void          pppd_end_chat(Account *account);
-void          pppd_kill(Account *account);
+gint          pppd_end_chat(Account *account);
+gint          pppd_device_number(Account *account, int *device_number);
 pid_t         pppd_pid(Account *account);
-gint          pppd_device_number(Account *account);
+gint          pppd_kill(Account *account);
 
 #endif /* __pppd_h__ */
