@@ -4,6 +4,10 @@
  * Author: Federico Mena
  */
 
+#ifndef GNOMELOCALEDIR
+#define GNOMELOCALEDIR "/usr/share/locale"
+#endif
+
 #include <config.h>
 #include <unistd.h>
 #include "gnome.h"
@@ -24,6 +28,9 @@ main(int argc, char **argv)
 {
 	poptContext ctx;
 	char **args;
+
+	bindtextdomain(PACKAGE, GNOMELOCALEDIR);
+	textdomain(PACKAGE);
 	
 	gnome_init_with_popt_table ("gnometalk", VERSION, argc, argv, options, 0, &ctx);
 	args = poptGetArgs (ctx);
